#!/usr/bin/env python
#
# Trigger a CI job on a gitlab project repository from the command-line.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import os.path  as op
import textwrap as tw
import             sys
import             argparse
import             datetime


from fsl_ci_utils.conda_api import (get_project_jobs,
                                    trigger_job,
                                    parse_gitlab_date)


SERVER_URL = 'https://git.fmrib.ox.ac.uk'
"""Default gitlab instance URL, if not specified on the command.line."""


JOB_NAME = 'deploy-conda-package'
"""Default name of CI job to trigger."""


def parseArgs():
    name   = op.basename(__file__)
    usage  = f'Usage: {name} -t <token> [options] recipe_path'
    desc  = tw.dedent(f"""
    This script looks up the most recently submitted CI job on the
    specified FSL conda recipe repository which has a specific name
    (default {JOB_NAME}), and has not yet been executed, and
    starts it running.
    """).strip()

    helps  = {
        'recipe_path' :
        'Path (namespace/project) of the gitlab recipe repository',

        'token' :
        'Gitlab API access token with read+write access',

        'server' :
        f'Gitlab server (default: {SERVER_URL})',

        'age' :
        'Only consider jobs which are at most this '
        'old, in hours (default: 1)',

        'job' :
        f'Name of job to trigger (default: {JOB_NAME}).',
    }

    parser = argparse.ArgumentParser(
        usage=usage,
        description=desc,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('recipe_path',    help=helps['recipe_path'])
    parser.add_argument('-s', '--server', help=helps['server'],
                        default=SERVER_URL)
    parser.add_argument('-t', '--token',  help=helps['token'],
                        required=True)
    parser.add_argument('-a', '--age',    help=helps['age'],
                        type=int, default=1)
    parser.add_argument('-j', '--job',    help=helps['job'],
                        default=JOB_NAME)

    return parser.parse_args()


def find_latest_deploy_job(recipe_path, server, token, jobname, age):
    """Return info about the most recent ``deploy-conda-package`` CI job,
    which was submitted at most ``age`` hours ago.
    """
    now   = datetime.datetime.now().astimezone()
    jobs  = get_project_jobs(recipe_path, server, token, scope='manual')
    found = None

    for job in jobs:
        created  = parse_gitlab_date(job['created_at'])
        timediff = (now - created).total_seconds() / 3600

        if all((job['name'] == jobname, timediff <= age)):
            jid = job['id']
            print(f'Found job {jobname} [ID {jid}] '
                  f'scheduled {timediff} hours ago')

            found = job
            break

    return found

def main():
    args = parseArgs()
    job  = find_latest_deploy_job(args.recipe_path,
                                  args.server,
                                  args.token,
                                  args.job,
                                  args.age)

    if job is None:
        print(f'Cannot find a CI job on project {args.recipe_path} '
              f'with name {args.job} - aborting.')
        sys.exit(1)

    trigger_job(args.recipe_path, job['id'], args.server, args.token)


if __name__ == '__main__':
    main()
