#!/usr/bin/env python
#
# plugin_manifest.py - Classes and functions for working with FSL plugin
#                      manifest files.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains classes and functions for working with FSL plugin
manifest files. A plugin manifest file is a JSON file which contains a
list of FSL plugin definitions, e.g.::

    [
        {
            "name"        : "Unix Introduction",
            "category"    : "fsl_course_data",
            "url"         : "http://.../UnixIntro.zip",
            "checksum"    : "f8222...",
            "description" : "Data for the unix introduction",
            "destination" : "~/"
        },
        {
            "name"        : "Preparatory material",
            "category"    : "fsl_course_data",
            "url"         : "http://...preCourse.zip",
            "checksum"    : "e659f",
            "description" : "Data for the preparatory practicals",
            "destination" : "~/"
        }
    ]

Each plugin definition must contain the ``url`` field; all other fieids are
optional but recommended:

  - ``name``:        A unique name for the plugin. Defaults to the filename
                     component of the ``url``.
  - ``url``:         URL of the plugin archive file.
  - ``category``:    The plugin category, used to organise plugins.
  - ``checksum``:    A SHA256 checksum of the plugin archive file, used to
                     verify that it has been successfully downloaded.
  - ``description``: An extended description of the plugin
  - ``destination``: Default installation directory. May contain environment
                     variables and ``~``.
"""


import os.path        as op
import                   os
import                   json
import                   logging
import                   pathlib
import                   tempfile
import                   dataclasses
import urllib.parse   as urlparse
import urllib.request as urlrequest

from typing import Union, Optional, List

import fsl.add_module.routines as routines


log = logging.getLogger(__name__)


class ManifestInvalid(Exception):
    """Exception raised by the :func:`downloadManifest` function if a
    manifest file appears to be invalid.
    """


UNCATEGORISED = 'uncategorised'
"""Category label used for otherwise uncategorised plugins."""


@dataclasses.dataclass
class Plugin:
    """Simple data class which represents a single entry in a FSL plugin
    manifest file.
    """

    name : str
    """Name used to uniquely identify this plugin."""

    url : Union[str, pathlib.Path]
    """URL of the plugin archive file. """

    category : str = UNCATEGORISED
    """Plugin category - plugins may be grouped by category."""

    description : Optional[str] = None
    """Description of the plugin. """

    checksum : Optional[str] = None
    """SHA256 checksum of the plugin archive file, used to verify downloads.
    """

    destination : Union[None, str, pathlib.Path] = None
    """Installation/destination directory - where the plugin archive file
    will be extracted. May have a default value in the manifest, but may
    be overridden at user request. Any environment variables and the ``~``
    character will be replaced with the variable value/user home directory.
    """

    origDestination : Union[None, str, pathlib.Path] = None
    """Original value of ``destination``, before environment variable/tilde
    expansion.

    If the original destination contains any environment variables that
    are not set, the user is warned.
    """

    archiveFile : Union[None, str, pathlib.Path] = None
    """Path to downloaded archive file (manually assigned once the file has
    been downloaded).
    """

    checksumPassed : Optional[bool] = None
    """Flag indicating whether the checksum of the downloaded file matched
    the checksum in the manifest (manually assigned once the file has been
    downloaded).
    """


class Manifest(dict):
    """The ``Manifest``class simply a container for :class:`Plugin` instances.
    A ``Manifest`` is a dict containing ``{name : Plugin}`` mappings.

    The :meth:`addPlugin` method may be used to add a new :class:`Plugin` to
    the manifest
    """


    def addPlugin(self,
                  url         : Union[str, pathlib.Path],
                  name        : Optional[str] = None,
                  description : Optional[str] = None,
                  category    : Optional[str] = None,
                  checksum    : Optional[str] = None,
                  destination : Optional[str] = None) -> str:
        """Add a new :class:`Plugin` to the manifest.

        :arg url:         URL of plugin archive file.
        :arg name:        Plugin name. Defaults to filename component of
                          ``url``.
        :arg description: Extended description of plugin
        :arg category:    Plugin category
        :arg checksum:    SHA256 checksum of archive file.
        :arg destination: Default installation directory.
        :returns:         The registered plugin name (equivalent to ``name``,
                          if provided).
        """

        # support paths to local files
        if op.exists(url):
            url         = routines.expand(url)
            archiveFile = url
            url         = urlparse.urljoin(
                'file:', urlrequest.pathname2url(url))
        else:
            archiveFile = None

        # name defaults to filename in URL
        if name is None:
            name = op.basename(urlparse.urlparse(url).path)

        if category is None:
            category = UNCATEGORISED

        # install location defaults to $FSLDIR
        # if set or, failing that, to pwd
        if destination is None:
            destination = os.environ.get('FSLDIR', os.getcwd())

        # allow env vars and tilde in
        # JSON destination field
        origDestination = destination
        destination     = routines.expand(destination)

        if name in self:
            log.warning('Module with name %s [%s] already exists! Overwriting '
                        'with [%s]', name, self[name].url, url)

        self[name] = Plugin(
            name=name,
            url=url,
            category=category,
            description=description,
            checksum=checksum,
            destination=destination,
            origDestination=origDestination,
            archiveFile=archiveFile,
            checksumPassed=None)

        log.debug('Registered plugin" %s', self[name])

        return name


    @property
    def categories(self) -> List[str]:
        """Returns a list of all plugin categories in the manifest. """
        cats = [p.category for p in self.values()]
        return list(sorted(set(cats)))


    def getCategory(self, category : str) -> List[Plugin]:
        """Returns a list of all available plugins in the specified
        ``category``.
        """
        return list([p for p in self.values() if p.category == category])


def downloadManifest(url : Union[str, pathlib.Path]) -> Manifest:
    """Downloads a FSL plugin manifest from the given ``url``.

    Raises a :exc:`ManifestInvalid` error if the manifest file appears to
    be invalid in any way.

    :arg url: URL to the manifest file
    :returns: A :class:`Manifest` object.
    """
    log.info('Downloading manifest file from %s...', url)

    # allow local paths in addition
    # to urls for the manifest file
    if op.exists(url):
        url = urlparse.urljoin(
            'file:', urlrequest.pathname2url(op.abspath(url)))

    manifest = Manifest()

    with tempfile.TemporaryDirectory() as d:
        dest = op.join(d, 'manifest.txt')
        routines.downloadFile(url, dest, progress=False)
        rawmanifest = open(dest, 'rt').read()

    try:
        rawmanifest = json.loads(rawmanifest)
    except json.decoder.JSONDecodeError as e:
        raise ManifestInvalid(f'The manifest file {url} '
                              'is not a valid JSON file') from e

    for plugin in rawmanifest:

        # plugins must at least contain a URL
        try:
            manifest.addPlugin(
                url=plugin['url'],
                name=plugin.get('name'),
                description=plugin.get('description'),
                category=plugin.get('category'),
                checksum=plugin.get('checksum'),
                destination=plugin.get('destination'))
        except KeyError as e:
            raise ManifestInvalid(f'The manifest file {url} '
                                  'contains an invaluid module '
                                  'definition: {plugin}') from e

    return manifest
